# include<iostream>
#include "window.h"
#include "glerror.h"


static const char* kVertexShaderFile="Shader.vertex.glsl";
static const char* kFragmentShaderFile="Shader.fragment.glsl";
static const char* kTessControlShaderFile="Shader.tess.control.glsl";
static const char* kTessEvalShaderFile="Shader.tess.eval.glsl";

const Material  kMaterial={
    {0.6f, 0.6f, 0.6f, 1.0f}, //Ambient
    {1.0f, 1.0f, 0.0f, 1.0f}, //Diffuse
    {0.6f, 0.6f, 0.6f, 1.0f}, //Specular
    {0.0f, 0.0f, 0.0f, 1.0f}, //Emission
    60.0f //shininess
  };


const PointLight kPointLight={
    {-5.0f, 8.0f, 5.0f, 1.0f}, //position
    {0.3f, 0.3f, 0.3f, 1.0f}, //ambient
    {1.0f, 1.0f, 1.0f, 1.0f}, //diffuse
    {1.0f, 1.0f, 1.0f, 1.0f}, //specular
    {0.5f, 0.005f, 0.0125f}	//attenuation
};

const GLfloat kBezierOne[16][4] ={
    {-2, 1, 0, 1},
        {-2.0f/3.0f, 1.0f/3.0f, 4.0f/3.0f, 1.0f/3.0f},
        { 2.0f/3.0f, 1.0f/3.0f, 4.0f/3.0f, 1.0f/3.0f},
    {2, 1, 0, 1},
    {-1.5, 0.5, 0,1},
        {-0.5, 1.0f/6.0f, 1, 1.0f/3.0f},
        { 0.5, 1.0f/6.0f, 1, 1.0f/3.0f},
    {1.5, 0.5, 0, 1},
    {-3, 0, 0,1},
        {-1, 0, 2, 1.0f/3.0f},
        { 1, 0, 2, 1.0f/3.0f},
    {3, 0, 0, 1},
    {-2, -1, 0, 1},
        {-2.0f/3.0f, -1.0f/3.0f, 4.0f/3.0f, 1.0f/3.0f},
        { 2.0f/3.0f, -1.0f/3.0f, 4.0f/3.0f, 1.0f/3.0f},
    {2, -1, 0, 1}
};


Window::Window(const char * title, int width, int height){
    title_ = title;
    width_ = width;
    height_ = height;
    last_time_ = 0;
}

void Window::Initialize(int major_gl_version, int minor_gl_version){

    InitGlfwOrDie(major_gl_version, minor_gl_version);
    InitGlewOrDie();


    std::cout << "OpenGL initialized: OpenGL version: " << glGetString(GL_VERSION) << " GLSL version: " << glGetString(GL_SHADING_LANGUAGE_VERSION) << std::endl;

    InitModels();
    InitPrograms();

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glClearColor(1.0f, 1.0f, 1.0f, 1.0f);

}


void Window::InitGlfwOrDie(int major_gl_version, int minor_gl_version){
    if ( !glfwInit() ) {
        std::cerr << "ERROR: Could not initialize GLFW" << std::endl;
        exit(EXIT_FAILURE);
    }


    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, major_gl_version);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, minor_gl_version);
    glfwWindowHint(GLFW_SAMPLES, 8);
    glEnable(GL_MULTISAMPLE);

#ifdef DEBUG
    glfwWindowHint(GLFW_OPENGL_DEBUG_CONTEXT, GLFW_TRUE);
#endif
    window_ = glfwCreateWindow(width_, height_, title_, nullptr, nullptr);
    if (!window_){
        std::cerr << "ERROR: Could not create a new rendering window" << std::endl;
        glfwTerminate();
        exit(EXIT_FAILURE);
    }
    glfwMakeContextCurrent(window_);
}

void Window::InitGlewOrDie(){
    GLenum glew_init_result;
    glewExperimental = GL_TRUE;
    glew_init_result = glewInit();

    if (GLEW_OK != glew_init_result) {
        std::cerr << "Glew ERROR: " << glewGetErrorString(glew_init_result) << std::endl;
        glfwTerminate();
        exit(EXIT_FAILURE);
    }
#ifdef DEBUG
    if(glDebugMessageCallback){
        std::cout << "Register OpenGL debug callback " << std::endl;
        glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS);
        glDebugMessageCallback((GLDEBUGPROC) OpenglCallbackFunction, NULL);
        GLuint unused_ids = 0;
        glDebugMessageControl(GL_DONT_CARE,
            GL_DONT_CARE,
            GL_DONT_CARE,
            0,
            &unused_ids,
            GL_FALSE);
    }
    else
        std::cout << "glDebugMessageCallback not available" << std::endl;
#endif
}


void Window::InitModels(){
    bezier_.Initialize(kBezierOne);
    bezier_.SetMaterial(kMaterial);
}

void Window::InitPrograms(){
    view_matrix_.Translate(0, 0, -10);
    projection_matrix_ = Mat4::CreatePerspectiveProjectionMatrix(45, (float)width_/(float)height_, 0.1f, 100.0f);

    program_.Initialize(kVertexShaderFile, kFragmentShaderFile, kTessControlShaderFile, kTessEvalShaderFile);
    glUseProgram(program_);
    program_.SetLight(kPointLight);
    program_.SetProjectionMatrix(projection_matrix_);
    program_.SetViewMatrix(view_matrix_);

    glUseProgram(0);
}


void Window::SetViewMatrix(){
    glUseProgram(program_);
    program_.SetViewMatrix(view_matrix_);
    glUseProgram(0);
}

void Window::SetProjectionMatrix(){
    glUseProgram(program_);
    program_.SetProjectionMatrix(projection_matrix_);
    glUseProgram(0);
}

void Window::Resize(int new_width, int new_height){
    width_ = new_width;
    height_ = new_height;
    projection_matrix_ = Mat4::CreatePerspectiveProjectionMatrix(45, (float)width_/(float)height_, 0.1f, 100.0f);
    SetProjectionMatrix();
    glViewport(0, 0, width_, height_);
}



void Window::KeyEvent(int key, int /*scancode*/, int action, int /*mods*/){
    if(action == GLFW_PRESS){
        switch (key){
            case GLFW_KEY_ESCAPE:
                glfwSetWindowShouldClose(window_, GLFW_TRUE);
            break;
            case GLFW_KEY_LEFT:
              bezier_.SlowDown();
            break;
            case GLFW_KEY_RIGHT:
              bezier_.SpeedUp();
            break;
            case GLFW_KEY_SPACE:
              bezier_.ToggleAnimated();
            break;
        }
    }
    else if(action == GLFW_REPEAT){
        switch (key){
            case GLFW_KEY_LEFT:
              bezier_.SlowDown();
            break;
            case GLFW_KEY_RIGHT:
              bezier_.SpeedUp();
            break;
            default:
            break;
        }
    }
}


void Window::Run(void){
    while (!glfwWindowShouldClose(window_)){
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        clock_t now = clock();
        if (last_time_ == 0) last_time_ = now;
        GLfloat delta_t = (float)(now - last_time_) / CLOCKS_PER_SEC;
        bezier_.Move(delta_t);
        last_time_ = now;

        bezier_.Draw(program_);

        glfwSwapBuffers(window_);
        glfwPollEvents();
    }

}
